@@ -10,22 +10,29 @@
 #include "miscadmin.h"
 
 #include "access/htup_details.h"
+#include "access/sysattr.h"
+#include "access/xact.h"
 #include "catalog/catalog.h"
 #include "catalog/index.h"
 #include "catalog/namespace.h"
 #include "commands/defrem.h"
 #include "commands/tablecmds.h"
+#include "distributed/citus_ruleutils.h"
 #include "distributed/master_protocol.h"
 #include "distributed/metadata_cache.h"
+#include "distributed/multi_copy.h"
 #include "distributed/multi_utility.h"
 #include "distributed/multi_join_order.h"
 #include "distributed/transmit.h"
 #include "distributed/worker_manager.h"
 #include "distributed/worker_protocol.h"
+#include "foreign/foreign.h"
+#include "executor/executor.h"
 #include "parser/parser.h"
 #include "parser/parse_utilcmd.h"
 #include "storage/lmgr.h"
 #include "tcop/pquery.h"
+#include "tcop/utility.h"
 #include "utils/builtins.h"
 #include "utils/inval.h"
 #include "utils/lsyscache.h"
@@ -50,7 +57,8 @@ static bool IsTransmitStmt(Node *parsetree);
 static void VerifyTransmitStmt(CopyStmt *copyStatement);
 
 /* Local functions forward declarations for processing distributed table commands */
-static Node * ProcessCopyStmt(CopyStmt *copyStatement);
+static Node * ProcessCopyStmt(CopyStmt *copyStatement, char *completionTag,
+							  bool *commandMustRunAsOwner);
 static Node * ProcessIndexStmt(IndexStmt *createIndexStatement,
 							   const char *createIndexCommand);
 static Node * ProcessDropIndexStmt(DropStmt *dropIndexStatement,
@@ -65,7 +73,7 @@ static void ErrorIfUnsupportedAlterTableStmt(AlterTableStmt *alterTableStatement
 static void ErrorIfDistributedRenameStmt(RenameStmt *renameStatement);
 
 /* Local functions forward declarations for helper functions */
-static void WarnIfDropCitusExtension(DropStmt *dropStatement);
+static void CreateLocalTable(RangeVar *relation, char *nodeName, int32 nodePort);
 static bool IsAlterTableRenameStmt(RenameStmt *renameStatement);
 static void ExecuteDistributedDDLCommand(Oid relationId, const char *ddlCommandString);
 static bool ExecuteCommandOnWorkerShards(Oid relationId, const char *commandString,
@@ -73,6 +81,8 @@ static bool ExecuteCommandOnWorkerShards(Oid relationId, const char *commandStri
 static bool AllFinalizedPlacementsAccessible(Oid relationId);
 static void RangeVarCallbackForDropIndex(const RangeVar *rel, Oid relOid, Oid oldRelOid,
 										 void *arg);
+static void CheckCopyPermissions(CopyStmt *copyStatement);
+static List * CopyGetAttnums(TupleDesc tupDesc, Relation rel, List *attnamelist);
 
 
 /*
@@ -95,6 +105,10 @@ multi_ProcessUtility(Node *parsetree,
 					 DestReceiver *dest,
 					 char *completionTag)
 {
+	bool commandMustRunAsOwner = false;
+	Oid savedUserId = InvalidOid;
+	int savedSecurityContext = 0;
+
 	/*
 	 * TRANSMIT used to be separate command, but to avoid patching the grammar
 	 * it's no overlaid onto COPY, but with FORMAT = 'transmit' instead of the
@@ -122,7 +136,13 @@ multi_ProcessUtility(Node *parsetree,
 
 	if (IsA(parsetree, CopyStmt))
 	{
-		parsetree = ProcessCopyStmt((CopyStmt *) parsetree);
+		parsetree = ProcessCopyStmt((CopyStmt *) parsetree, completionTag,
+									&commandMustRunAsOwner);
+
+		if (parsetree == NULL)
+		{
+			return;
+		}
 	}
 
 	if (IsA(parsetree, IndexStmt))
@@ -137,10 +157,6 @@ multi_ProcessUtility(Node *parsetree,
 		{
 			parsetree = ProcessDropIndexStmt(dropStatement, queryString);
 		}
-		else if (dropStatement->removeType == OBJECT_EXTENSION)
-		{
-			WarnIfDropCitusExtension(dropStatement);
-		}
 	}
 
 	if (IsA(parsetree, AlterTableStmt))
@@ -190,45 +206,25 @@ multi_ProcessUtility(Node *parsetree,
 	}
 	else if (IsA(parsetree, CreateRoleStmt) && CitusHasBeenLoaded())
 	{
-		ereport(NOTICE, (errmsg("Citus does not support CREATE ROLE/USER "
-								"for distributed databases"),
-						 errdetail("Multiple roles are currently supported "
-								   "only for local tables")));
+		ereport(NOTICE, (errmsg("not propagating CREATE ROLE/USER commands to worker"
+								" nodes"),
+						 errhint("Connect to worker nodes directly to manually create all"
+								 " necessary users and roles.")));
+	}
+
+	if (commandMustRunAsOwner)
+	{
+		GetUserIdAndSecContext(&savedUserId, &savedSecurityContext);
+		SetUserIdAndSecContext(CitusExtensionOwner(), SECURITY_LOCAL_USERID_CHANGE);
 	}
 
 	/* now drop into standard process utility */
 	standard_ProcessUtility(parsetree, queryString, context,
 							params, dest, completionTag);
-}
 
-
-/*
- * WarnIfDropCitusExtension prints a WARNING if dropStatement includes dropping
- * citus extension.
- */
-static void
-WarnIfDropCitusExtension(DropStmt *dropStatement)
-{
-	ListCell *dropStatementObject = NULL;
-
-	Assert(dropStatement->removeType == OBJECT_EXTENSION);
-
-	foreach(dropStatementObject, dropStatement->objects)
+	if (commandMustRunAsOwner)
 	{
-		List *objectNameList = lfirst(dropStatementObject);
-		char *objectName = NameListToString(objectNameList);
-
-		/* we're only concerned with the citus extension */
-		if (strncmp("citus", objectName, NAMEDATALEN) == 0)
-		{
-			/*
-			 * Warn the user about the possibility of invalid cache. Also, see
-			 * function comment of CachedRelationLookup().
-			 */
-			ereport(WARNING, (errmsg("could not clean the metadata cache on "
-									 "DROP EXTENSION command"),
-							  errhint("Reconnect to the server again.")));
-		}
+		SetUserIdAndSecContext(savedUserId, savedSecurityContext);
 	}
 }
 
@@ -297,28 +293,20 @@ VerifyTransmitStmt(CopyStmt *copyStatement)
 
 /*
  * ProcessCopyStmt handles Citus specific concerns for COPY like supporting
- * COPYing from distributed tables and preventing unsupported actions.
+ * COPYing from distributed tables and preventing unsupported actions. The
+ * function returns a modified COPY statement to be executed, or NULL if no
+ * further processing is needed.
+ *
+ * commandMustRunAsOwner is an output parameter used to communicate to the caller whether
+ * the copy statement should be executed using elevated privileges. If
+ * ProcessCopyStmt that is required, a call to CheckCopyPermissions will take
+ * care of verifying the current user's permissions before ProcessCopyStmt
+ * returns.
  */
 static Node *
-ProcessCopyStmt(CopyStmt *copyStatement)
+ProcessCopyStmt(CopyStmt *copyStatement, char *completionTag, bool *commandMustRunAsOwner)
 {
-	/*
-	 * We first check if we have a "COPY (query) TO filename". If we do, copy doesn't
-	 * accept relative file paths. However, SQL tasks that get assigned to worker nodes
-	 * have relative paths. We therefore convert relative paths to absolute ones here.
-	 */
-	if (copyStatement->relation == NULL &&
-		!copyStatement->is_from &&
-		!copyStatement->is_program &&
-		copyStatement->filename != NULL)
-	{
-		const char *filename = copyStatement->filename;
-
-		if (!is_absolute_path(filename) && JobDirectoryElement(filename))
-		{
-			copyStatement->filename = make_absolute_path(filename);
-		}
-	}
+	*commandMustRunAsOwner = false; /* make sure variable is initialized */
 
 	/*
 	 * We check whether a distributed relation is affected. For that, we need to open the
@@ -327,26 +315,54 @@ ProcessCopyStmt(CopyStmt *copyStatement)
 	 */
 	if (copyStatement->relation != NULL)
 	{
-		Relation copiedRelation = NULL;
 		bool isDistributedRelation = false;
+		bool isCopyFromWorker = IsCopyFromWorker(copyStatement);
 
-		copiedRelation = heap_openrv(copyStatement->relation, AccessShareLock);
+		if (isCopyFromWorker)
+		{
+			RangeVar *relation = copyStatement->relation;
+			NodeAddress *masterNodeAddress = MasterNodeAddress(copyStatement);
+			char *nodeName = masterNodeAddress->nodeName;
+			int32 nodePort = masterNodeAddress->nodePort;
 
-		isDistributedRelation = IsDistributedTable(RelationGetRelid(copiedRelation));
+			CreateLocalTable(relation, nodeName, nodePort);
 
-		/* ensure future lookups hit the same relation */
-		copyStatement->relation->schemaname = get_namespace_name(
-			RelationGetNamespace(copiedRelation));
+			/*
+			 * We expect copy from worker to be on a distributed table; otherwise,
+			 * it fails in CitusCopyFrom() while checking the partition method.
+			 */
+			isDistributedRelation = true;
+		}
+		else
+		{
+			bool isFrom = copyStatement->is_from;
+			Relation copiedRelation = NULL;
 
-		heap_close(copiedRelation, NoLock);
+			/* consider using RangeVarGetRelidExtended to check perms before locking */
+			copiedRelation = heap_openrv(copyStatement->relation,
+										 isFrom ? RowExclusiveLock : AccessShareLock);
+
+			isDistributedRelation = IsDistributedTable(RelationGetRelid(copiedRelation));
+
+			/* ensure future lookups hit the same relation */
+			copyStatement->relation->schemaname = get_namespace_name(
+				RelationGetNamespace(copiedRelation));
+
+			heap_close(copiedRelation, NoLock);
+		}
 
 		if (isDistributedRelation)
 		{
 			if (copyStatement->is_from)
 			{
-				ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-								errmsg("cannot execute COPY FROM on a distributed table "
-									   "on master node")));
+				/* check permissions, we're bypassing postgres' normal checks */
+				if (!isCopyFromWorker)
+				{
+					CheckCopyPermissions(copyStatement);
+				}
+
+				CitusCopyFrom(copyStatement, completionTag);
+				return NULL;
 			}
 			else if (!copyStatement->is_from)
 			{
@@ -378,6 +394,50 @@ ProcessCopyStmt(CopyStmt *copyStatement)
 		}
 	}
 
+
+	if (copyStatement->filename != NULL && !copyStatement->is_program)
+	{
+		const char *filename = copyStatement->filename;
+
+		if (CacheDirectoryElement(filename))
+		{
+			/*
+			 * Only superusers are allowed to copy from a file, so we have to
+			 * become superuser to execute copies to/from files used by citus'
+			 * query execution.
+			 *
+			 * XXX: This is a decidedly suboptimal solution, as that means
+			 * that triggers, input functions, etc. run with elevated
+			 * privileges.  But this is better than not being able to run
+			 * queries as normal user.
+			 */
+			*commandMustRunAsOwner = true;
+
+			/*
+			 * Have to manually check permissions here as the COPY is will be
+			 * run as a superuser.
+			 */
+			if (copyStatement->relation != NULL)
+			{
+				CheckCopyPermissions(copyStatement);
+			}
+
+			/*
+			 * Check if we have a "COPY (query) TO filename". If we do, copy
+			 * doesn't accept relative file paths. However, SQL tasks that get
+			 * assigned to worker nodes have relative paths. We therefore
+			 * convert relative paths to absolute ones here.
+			 */
+			if (copyStatement->relation == NULL &&
+				!copyStatement->is_from &&
+				!is_absolute_path(filename))
+			{
+				copyStatement->filename = make_absolute_path(filename);
+			}
+		}
+	}
+
+
 	return (Node *) copyStatement;
 }
 
@@ -416,6 +476,12 @@ ProcessIndexStmt(IndexStmt *createIndexStatement, const char *createIndexCommand
 			lockmode = ShareUpdateExclusiveLock;
 		}
 
+		/*
+		 * XXX: Consider using RangeVarGetRelidExtended with a permission
+		 * checking callback. Right now we'll acquire the lock before having
+		 * checked permissions, and will only fail when executing the actual
+		 * index statements.
+		 */
 		relation = heap_openrv(createIndexStatement->relation, lockmode);
 		relationId = RelationGetRelid(relation);
 
@@ -565,6 +631,25 @@ ProcessAlterTableStmt(AlterTableStmt *alterTableStatement, const char *alterTabl
 static void
 ErrorIfUnsupportedIndexStmt(IndexStmt *createIndexStatement)
 {
+	Oid namespaceId;
+	Oid indexRelationId;
+	char *indexRelationName = createIndexStatement->idxname;
+
+	if (indexRelationName == NULL)
+	{
+		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						errmsg("creating index without a name on a distributed table is "
+							   "currently unsupported")));
+	}
+
+	namespaceId = get_namespace_oid(createIndexStatement->relation->schemaname, false);
+	indexRelationId = get_relname_relid(indexRelationName, namespaceId);
+	if (indexRelationId != InvalidOid)
+	{
+		ereport(ERROR, (errcode(ERRCODE_DUPLICATE_TABLE),
+						errmsg("relation \"%s\" already exists", indexRelationName)));
+	}
+
 	if (createIndexStatement->tableSpace != NULL)
 	{
 		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
@@ -572,19 +657,61 @@ ErrorIfUnsupportedIndexStmt(IndexStmt *createIndexStatement)
 							   "currently unsupported")));
 	}
 
-	if (createIndexStatement->unique)
-	{
-		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-						errmsg("creating unique indexes on distributed tables is "
-							   "currently unsupported")));
-	}
-
 	if (createIndexStatement->concurrent)
 	{
 		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 						errmsg("creating indexes concurrently on distributed tables is "
 							   "currently unsupported")));
 	}
+
+	if (createIndexStatement->unique)
+	{
+		RangeVar *relation = createIndexStatement->relation;
+		bool missingOk = false;
+
+		/* caller uses ShareLock for non-concurrent indexes, use the same lock here */
+		LOCKMODE lockMode = ShareLock;
+		Oid relationId = RangeVarGetRelid(relation, lockMode, missingOk);
+		Var *partitionKey = PartitionKey(relationId);
+		char partitionMethod = PartitionMethod(relationId);
+		List *indexParameterList = NIL;
+		ListCell *indexParameterCell = NULL;
+		bool indexContainsPartitionColumn = false;
+
+		if (partitionMethod == DISTRIBUTE_BY_APPEND)
+		{
+			ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+							errmsg("creating unique indexes on append-partitioned tables "
+								   "is currently unsupported")));
+		}
+
+		indexParameterList = createIndexStatement->indexParams;
+		foreach(indexParameterCell, indexParameterList)
+		{
+			IndexElem *indexElement = (IndexElem *) lfirst(indexParameterCell);
+			char *columnName = indexElement->name;
+			AttrNumber attributeNumber = InvalidAttrNumber;
+
+			/* column name is null for index expressions, skip it */
+			if (columnName == NULL)
+			{
+				continue;
+			}
+
+			attributeNumber = get_attnum(relationId, columnName);
+			if (attributeNumber == partitionKey->varattno)
+			{
+				indexContainsPartitionColumn = true;
+			}
+		}
+
+		if (!indexContainsPartitionColumn)
+		{
+			ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+							errmsg("creating unique indexes on non-partition "
+								   "columns is currently unsupported")));
+		}
+	}
 }
 
 
@@ -735,6 +862,82 @@ ErrorIfDistributedRenameStmt(RenameStmt *renameStatement)
 }
 
 
+/*
+ * CreateLocalTable gets DDL commands from the remote node for the given
+ * relation. Then, it creates the local relation as temporary and on commit drop.
+ */
+static void
+CreateLocalTable(RangeVar *relation, char *nodeName, int32 nodePort)
+{
+	List *ddlCommandList = NIL;
+	ListCell *ddlCommandCell = NULL;
+
+	char *relationName = relation->relname;
+	char *schemaName = relation->schemaname;
+	char *qualifiedName = quote_qualified_identifier(schemaName, relationName);
+
+	/* fetch the ddl commands needed to create the table */
+	StringInfo tableNameStringInfo = makeStringInfo();
+	appendStringInfoString(tableNameStringInfo, qualifiedName);
+
+	/*
+	 * The warning message created in TableDDLCommandList() is descriptive
+	 * enough; therefore, we just throw an error which says that we could not
+	 * run the copy operation.
+	 */
+	ddlCommandList = TableDDLCommandList(nodeName, nodePort, tableNameStringInfo);
+	if (ddlCommandList == NIL)
+	{
+		ereport(ERROR, (errmsg("could not run copy from the worker node")));
+	}
+
+	/* apply DDL commands against the local database */
+	foreach(ddlCommandCell, ddlCommandList)
+	{
+		StringInfo ddlCommand = (StringInfo) lfirst(ddlCommandCell);
+		Node *ddlCommandNode = ParseTreeNode(ddlCommand->data);
+		bool applyDDLCommand = false;
+
+		if (IsA(ddlCommandNode, CreateStmt) ||
+			IsA(ddlCommandNode, CreateForeignTableStmt))
+		{
+			CreateStmt *createStatement = (CreateStmt *) ddlCommandNode;
+
+			/* create the local relation as temporary and on commit drop */
+			createStatement->relation->relpersistence = RELPERSISTENCE_TEMP;
+			createStatement->oncommit = ONCOMMIT_DROP;
+
+			/* temporarily strip schema name */
+			createStatement->relation->schemaname = NULL;
+
+			applyDDLCommand = true;
+		}
+		else if (IsA(ddlCommandNode, CreateForeignServerStmt))
+		{
+			CreateForeignServerStmt *createServerStmt =
+				(CreateForeignServerStmt *) ddlCommandNode;
+			if (GetForeignServerByName(createServerStmt->servername, true) == NULL)
+			{
+				/* create server if not exists */
+				applyDDLCommand = true;
+			}
+		}
+		else if ((IsA(ddlCommandNode, CreateExtensionStmt)))
+		{
+			applyDDLCommand = true;
+		}
+
+		/* run only a selected set of DDL commands */
+		if (applyDDLCommand)
+		{
+			ProcessUtility(ddlCommandNode, CreateCommandTag(ddlCommandNode),
+						   PROCESS_UTILITY_TOPLEVEL, NULL, None_Receiver, NULL);
+			CommandCounterIncrement();
+		}
+	}
+}
+
+
 /*
  * IsAlterTableRenameStmt returns true if the passed in RenameStmt operates on a
  * distributed table or its objects. This includes:
@@ -859,6 +1062,7 @@ ExecuteCommandOnWorkerShards(Oid relationId, const char *commandString,
 	bool isFirstPlacement = true;
 	ListCell *shardCell = NULL;
 	List *shardList = NIL;
+	char *relationOwner = TableOwner(relationId);
 
 	shardList = LoadShardList(relationId);
 	foreach(shardCell, shardList)
@@ -882,7 +1086,7 @@ ExecuteCommandOnWorkerShards(Oid relationId, const char *commandString,
 			uint32 workerPort = placement->nodePort;
 
 			List *queryResultList = ExecuteRemoteQuery(workerName, workerPort,
-													   applyCommand);
+													   relationOwner, applyCommand);
 			if (queryResultList == NIL)
 			{
 				/*
@@ -1056,3 +1260,302 @@ RangeVarCallbackForDropIndex(const RangeVar *rel, Oid relOid, Oid oldRelOid, voi
 	}
 	/* *INDENT-ON* */
 }
+
+
+/*
+ * Check whether the current user has the permission to execute a COPY
+ * statement, raise ERROR if not. In some cases we have to do this separately
+ * from postgres' copy.c, because we have to execute the copy with elevated
+ * privileges.
+ *
+ * Copied from postgres, where it's part of DoCopy().
+ */
+static void
+CheckCopyPermissions(CopyStmt *copyStatement)
+{
+	/* *INDENT-OFF* */
+	bool		is_from = copyStatement->is_from;
+	Relation	rel;
+	Oid			relid;
+	List	   *range_table = NIL;
+	TupleDesc	tupDesc;
+	AclMode		required_access = (is_from ? ACL_INSERT : ACL_SELECT);
+	List	   *attnums;
+	ListCell   *cur;
+	RangeTblEntry *rte;
+
+	rel = heap_openrv(copyStatement->relation,
+								 is_from ? RowExclusiveLock : AccessShareLock);
+
+	relid = RelationGetRelid(rel);
+
+	rte = makeNode(RangeTblEntry);
+	rte->rtekind = RTE_RELATION;
+	rte->relid = relid;
+	rte->relkind = rel->rd_rel->relkind;
+	rte->requiredPerms = required_access;
+	range_table = list_make1(rte);
+
+	tupDesc = RelationGetDescr(rel);
+
+	attnums = CopyGetAttnums(tupDesc, rel, copyStatement->attlist);
+	foreach(cur, attnums)
+	{
+		int			attno = lfirst_int(cur) - FirstLowInvalidHeapAttributeNumber;
+
+		if (is_from)
+		{
+#if (PG_VERSION_NUM >= 90500)
+			rte->insertedCols = bms_add_member(rte->insertedCols, attno);
+#else
+			rte->modifiedCols = bms_add_member(rte->modifiedCols, attno);
+#endif
+		}
+		else
+		{
+			rte->selectedCols = bms_add_member(rte->selectedCols, attno);
+		}
+	}
+
+	ExecCheckRTPerms(range_table, true);
+
+	/* TODO: Perform RLS checks once supported */
+
+	heap_close(rel, NoLock);
+	/* *INDENT-ON* */
+}
+
+
+/* Helper for CheckCopyPermissions(), copied from postgres */
+static List *
+CopyGetAttnums(TupleDesc tupDesc, Relation rel, List *attnamelist)
+{
+	/* *INDENT-OFF* */
+	List	   *attnums = NIL;
+
+	if (attnamelist == NIL)
+	{
+		/* Generate default column list */
+		Form_pg_attribute *attr = tupDesc->attrs;
+		int			attr_count = tupDesc->natts;
+		int			i;
+
+		for (i = 0; i < attr_count; i++)
+		{
+			if (attr[i]->attisdropped)
+			{
+				continue;
+			}
+			attnums = lappend_int(attnums, i + 1);
+		}
+	}
+	else
+	{
+		/* Validate the user-supplied list and extract attnums */
+		ListCell   *l;
+
+		foreach(l, attnamelist)
+		{
+			char	   *name = strVal(lfirst(l));
+			int			attnum;
+			int			i;
+
+			/* Lookup column name */
+			attnum = InvalidAttrNumber;
+			for (i = 0; i < tupDesc->natts; i++)
+			{
+				if (tupDesc->attrs[i]->attisdropped)
+				{
+					continue;
+				}
+				if (namestrcmp(&(tupDesc->attrs[i]->attname), name) == 0)
+				{
+					attnum = tupDesc->attrs[i]->attnum;
+					break;
+				}
+			}
+			if (attnum == InvalidAttrNumber)
+			{
+				if (rel != NULL)
+				{
+					ereport(ERROR,
+							(errcode(ERRCODE_UNDEFINED_COLUMN),
+					errmsg("column \"%s\" of relation \"%s\" does not exist",
+						   name, RelationGetRelationName(rel))));
+				}
+				else
+				{
+					ereport(ERROR,
+							(errcode(ERRCODE_UNDEFINED_COLUMN),
+							 errmsg("column \"%s\" does not exist",
+									name)));
+				}
+			}
+			/* Check for duplicates */
+			if (list_member_int(attnums, attnum))
+			{
+				ereport(ERROR,
+						(errcode(ERRCODE_DUPLICATE_COLUMN),
+						 errmsg("column \"%s\" specified more than once",
+								name)));
+			}
+			attnums = lappend_int(attnums, attnum);
+		}
+	}
+
+	return attnums;
+	/* *INDENT-ON* */
+}
+
+
+/*
+ * ReplicateGrantStmt replicates GRANT/REVOKE command to worker nodes if the
+ * the statement affects distributed tables.
+ *
+ * NB: So far column level privileges are not supported.
+ */
+void
+ReplicateGrantStmt(Node *parsetree)
+{
+	GrantStmt *grantStmt = (GrantStmt *) parsetree;
+	StringInfoData privsString;
+	StringInfoData granteesString;
+	StringInfoData targetString;
+	StringInfoData ddlString;
+	ListCell *granteeCell = NULL;
+	ListCell *objectCell = NULL;
+	ListCell *privilegeCell = NULL;
+	bool isFirst = true;
+
+	initStringInfo(&privsString);
+	initStringInfo(&granteesString);
+	initStringInfo(&targetString);
+	initStringInfo(&ddlString);
+
+	/*
+	 * So far only table level grants are supported. Most other types of
+	 * grants aren't interesting anyway.
+	 */
+	if (grantStmt->targtype != ACL_TARGET_OBJECT ||
+		grantStmt->objtype != ACL_OBJECT_RELATION)
+	{
+		return;
+	}
+
+	/* deparse the privileges */
+	if (grantStmt->privileges == NIL)
+	{
+		appendStringInfo(&privsString, "ALL");
+	}
+	else
+	{
+		isFirst = true;
+		foreach(privilegeCell, grantStmt->privileges)
+		{
+			AccessPriv *priv = lfirst(privilegeCell);
+
+			if (!isFirst)
+			{
+				appendStringInfoString(&privsString, ", ");
+			}
+			isFirst = false;
+
+			Assert(priv->cols == NIL);
+			Assert(priv->priv_name != NULL);
+
+			appendStringInfo(&privsString, "%s", priv->priv_name);
+		}
+	}
+
+	/* deparse the privileges */
+	isFirst = true;
+	foreach(granteeCell, grantStmt->grantees)
+	{
+#if (PG_VERSION_NUM >= 90500)
+		RoleSpec *spec = lfirst(granteeCell);
+#else
+		PrivGrantee *spec = lfirst(granteeCell);
+#endif
+
+		if (!isFirst)
+		{
+			appendStringInfoString(&granteesString, ", ");
+		}
+		isFirst = false;
+
+#if (PG_VERSION_NUM >= 90500)
+		if (spec->roletype == ROLESPEC_CSTRING)
+		{
+			appendStringInfoString(&granteesString, quote_identifier(spec->rolename));
+		}
+		else if (spec->roletype == ROLESPEC_CURRENT_USER)
+		{
+			appendStringInfoString(&granteesString, "CURRENT_USER");
+		}
+		else if (spec->roletype == ROLESPEC_SESSION_USER)
+		{
+			appendStringInfoString(&granteesString, "SESSION_USER");
+		}
+		else if (spec->roletype == ROLESPEC_PUBLIC)
+		{
+			appendStringInfoString(&granteesString, "PUBLIC");
+		}
+#else
+		if (spec->rolname)
+		{
+			appendStringInfoString(&granteesString, quote_identifier(spec->rolname));
+		}
+		else
+		{
+			appendStringInfoString(&granteesString, "PUBLIC");
+		}
+#endif
+	}
+
+	/*
+	 * Deparse the target objects, and issue the deparsed statements to
+	 * workers, if applicable. That's so we easily can replicate statements
+	 * only to distributed relations.
+	 */
+	isFirst = true;
+	foreach(objectCell, grantStmt->objects)
+	{
+		RangeVar *relvar = (RangeVar *) lfirst(objectCell);
+		Oid relOid = RangeVarGetRelid(relvar, NoLock, false);
+		const char *grantOption = "";
+
+		if (!IsDistributedTable(relOid))
+		{
+			continue;
+		}
+
+		resetStringInfo(&targetString);
+		appendStringInfo(&targetString, "%s", generate_relation_name(relOid, NIL));
+
+		if (grantStmt->is_grant)
+		{
+			if (grantStmt->grant_option)
+			{
+				grantOption = " WITH GRANT OPTION";
+			}
+
+			appendStringInfo(&ddlString, "GRANT %s ON %s TO %s%s",
+							 privsString.data, targetString.data, granteesString.data,
+							 grantOption);
+		}
+		else
+		{
+			if (grantStmt->grant_option)
+			{
+				grantOption = "GRANT OPTION FOR ";
+			}
+
+			appendStringInfo(&ddlString, "REVOKE %s%s ON %s FROM %s",
+							 grantOption, privsString.data, targetString.data,
+							 granteesString.data);
+		}
+
+		ExecuteDistributedDDLCommand(relOid, ddlString.data);
+		resetStringInfo(&ddlString);
+	}
+}
