@@ -12,6 +12,7 @@
 
 #include "postgres.h"
 #include "funcapi.h"
+#include "miscadmin.h"
 
 #include "access/htup_details.h"
 #include "access/xact.h"
@@ -28,6 +29,8 @@
 #include "distributed/worker_manager.h"
 #include "nodes/makefuncs.h"
 #include "parser/scansup.h"
+#include "storage/lmgr.h"
+#include "utils/acl.h"
 #include "utils/builtins.h"
 #include "utils/datum.h"
 #include "utils/fmgroids.h"
@@ -42,10 +45,18 @@
 static uint64 * AllocateUint64(uint64 value);
 
 
+/* exports for SQL callable functions */
+PG_FUNCTION_INFO_V1(master_stage_shard_row);
+PG_FUNCTION_INFO_V1(master_stage_shard_placement_row);
+
+
 /*
  * LoadShardIntervalList returns a list of shard intervals related for a given
  * distributed table. The function returns an empty list if no shards can be
  * found for the given relation.
+ * Since LoadShardIntervalList relies on sortedShardIntervalArray, it returns
+ * a shard interval list whose elements are sorted on shardminvalue. Shard intervals
+ * with uninitialized shard min/max values are placed in the end of the list.
  */
 List *
 LoadShardIntervalList(Oid relationId)
@@ -59,7 +70,7 @@ LoadShardIntervalList(Oid relationId)
 		ShardInterval *newShardInterval = NULL;
 		newShardInterval = (ShardInterval *) palloc0(sizeof(ShardInterval));
 
-		CopyShardInterval(&cacheEntry->shardIntervalArray[i], newShardInterval);
+		CopyShardInterval(cacheEntry->sortedShardIntervalArray[i], newShardInterval);
 
 		shardList = lappend(shardList, newShardInterval);
 	}
@@ -68,9 +79,32 @@ LoadShardIntervalList(Oid relationId)
 }
 
 
+/*
+ * ShardIntervalCount returns number of shard intervals for a given distributed table.
+ * The function returns 0 if table is not distributed, or no shards can be found for
+ * the given relation id.
+ */
+int
+ShardIntervalCount(Oid relationId)
+{
+	DistTableCacheEntry *cacheEntry = DistributedTableCacheEntry(relationId);
+	int shardIntervalCount = 0;
+
+	if (cacheEntry->isDistributedTable)
+	{
+		shardIntervalCount = cacheEntry->shardIntervalArrayLength;
+	}
+
+	return shardIntervalCount;
+}
+
+
 /*
  * LoadShardList reads list of shards for given relationId from pg_dist_shard,
  * and returns the list of found shardIds.
+ * Since LoadShardList relies on sortedShardIntervalArray, it returns a shard
+ * list whose elements are sorted on shardminvalue. Shards with uninitialized
+ * shard min/max values are placed in the end of the list.
  */
 List *
 LoadShardList(Oid relationId)
@@ -81,7 +115,7 @@ LoadShardList(Oid relationId)
 
 	for (i = 0; i < cacheEntry->shardIntervalArrayLength; i++)
 	{
-		ShardInterval *currentShardInterval = &cacheEntry->shardIntervalArray[i];
+		ShardInterval *currentShardInterval = cacheEntry->sortedShardIntervalArray[i];
 		uint64 *shardIdPointer = AllocateUint64(currentShardInterval->shardId);
 
 		shardList = lappend(shardList, shardIdPointer);
@@ -585,3 +619,201 @@ BuildDistributionKeyFromColumnName(Relation distributedRelation, char *columnNam
 
 	return (Node *) column;
 }
+
+
+/*
+ * Check that the current user has `mode` permissions on relationId, error out
+ * if not. Superusers always have such permissions.
+ */
+void
+EnsureTablePermissions(Oid relationId, AclMode mode)
+{
+	AclResult aclresult;
+
+	aclresult = pg_class_aclcheck(relationId, GetUserId(), mode);
+
+	if (aclresult != ACLCHECK_OK)
+	{
+		aclcheck_error(aclresult, ACL_KIND_CLASS,
+					   get_rel_name(relationId));
+	}
+}
+
+
+/*
+ * Check that the current user has owner rights to relationId, error out if
+ * not. Superusers are regarded as owners.
+ */
+void
+EnsureTableOwner(Oid relationId)
+{
+	if (!pg_class_ownercheck(relationId, GetUserId()))
+	{
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_CLASS,
+					   get_rel_name(relationId));
+	}
+}
+
+
+/*
+ * Return a table's owner as a string.
+ */
+char *
+TableOwner(Oid relationId)
+{
+	Oid userId = InvalidOid;
+	HeapTuple tuple;
+
+	tuple = SearchSysCache1(RELOID, ObjectIdGetDatum(relationId));
+	if (!HeapTupleIsValid(tuple))
+	{
+		ereport(ERROR, (errcode(ERRCODE_UNDEFINED_TABLE),
+						errmsg("relation with OID %u does not exist", relationId)));
+	}
+
+	userId = ((Form_pg_class) GETSTRUCT(tuple))->relowner;
+
+	ReleaseSysCache(tuple);
+
+#if (PG_VERSION_NUM < 90500)
+	return GetUserNameFromId(userId);
+#else
+	return GetUserNameFromId(userId, false);
+#endif
+}
+
+
+/*
+ * master_stage_shard_row() inserts a row into pg_dist_shard, after performing
+ * basic permission checks.
+ *
+ * TODO: This function only exists for csql's \stage, and should not otherwise
+ * be used. Once \stage is removed, it'll be removed too.
+ */
+Datum
+master_stage_shard_row(PG_FUNCTION_ARGS)
+{
+	Oid distributedRelationId = InvalidOid;
+	uint64 shardId = 0;
+	char storageType = 0;
+	text *shardMinValue = NULL;
+	text *shardMaxValue = NULL;
+	Relation relation;
+
+	/*
+	 * Have to check arguments for NULLness as it can't be declared STRICT
+	 * because of min/max arguments, which have to be NULLable for new shards.
+	 */
+	if (PG_ARGISNULL(0))
+	{
+		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						errmsg("relation cannot be null")));
+	}
+	else if (PG_ARGISNULL(1))
+	{
+		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						errmsg("shard cannot be null")));
+	}
+	else if (PG_ARGISNULL(2))
+	{
+		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						errmsg("storage type cannot be null")));
+	}
+
+	distributedRelationId = PG_GETARG_OID(0);
+	shardId = PG_GETARG_INT64(1);
+	storageType = PG_GETARG_CHAR(2);
+
+	if (!PG_ARGISNULL(3))
+	{
+		shardMinValue = PG_GETARG_TEXT_P(3);
+	}
+
+	if (!PG_ARGISNULL(4))
+	{
+		shardMaxValue = PG_GETARG_TEXT_P(4);
+	}
+
+	relation = heap_open(distributedRelationId, RowExclusiveLock);
+
+	/*
+	 * Check permissions on relation. Note we require ACL_INSERT and not owner
+	 * rights - it'd be worse for security to require every user performing
+	 * data loads to be made a table owner - besides being more complex to set
+	 * up.
+	 */
+	EnsureTablePermissions(distributedRelationId, ACL_INSERT);
+
+	/* and finally actually insert the row */
+	InsertShardRow(distributedRelationId, shardId, storageType,
+				   shardMinValue, shardMaxValue);
+
+	heap_close(relation, NoLock);
+
+	PG_RETURN_VOID();
+}
+
+
+/*
+ * master_stage_shard_placement_row() inserts a row into
+ * pg_dist_shard_placment, after performing some basic checks.
+ *
+ * TODO: This function only exists for csql's \stage, and should not otherwise
+ * be used. Once \stage is removed, it'll be removed too.
+ */
+Datum
+master_stage_shard_placement_row(PG_FUNCTION_ARGS)
+{
+	uint64 shardId = PG_GETARG_INT64(0);
+	int32 shardState = PG_GETARG_INT32(1);
+	int32 shardLength = PG_GETARG_INT64(2);
+	char *nodeName = text_to_cstring(PG_GETARG_TEXT_P(3));
+	int nodePort = PG_GETARG_INT32(4);
+
+	Oid distributedRelationId = InvalidOid;
+	Relation relation = NULL;
+
+	Relation pgDistShard = heap_open(DistShardRelationId(), RowExclusiveLock);
+	ScanKeyData scanKey[1];
+	int scanKeyCount = 1;
+	SysScanDesc scanDescriptor = NULL;
+	HeapTuple heapTuple = NULL;
+
+	/* Lookup which table the shardid belongs to */
+	ScanKeyInit(&scanKey[0], Anum_pg_dist_shard_shardid,
+				BTEqualStrategyNumber, F_INT8EQ, Int64GetDatum(shardId));
+
+	scanDescriptor = systable_beginscan(pgDistShard,
+										DistShardShardidIndexId(), true,
+										NULL, scanKeyCount, scanKey);
+	heapTuple = systable_getnext(scanDescriptor);
+	if (HeapTupleIsValid(heapTuple))
+	{
+		Form_pg_dist_shard pgDistShardForm = (Form_pg_dist_shard) GETSTRUCT(heapTuple);
+		distributedRelationId = pgDistShardForm->logicalrelid;
+	}
+	else
+	{
+		ereport(ERROR, (errmsg("could not find valid entry for shard "
+							   UINT64_FORMAT, shardId)));
+	}
+	systable_endscan(scanDescriptor);
+
+	relation = heap_open(distributedRelationId, RowExclusiveLock);
+
+	/*
+	 * Check permissions on relation. Note we require ACL_INSERT and not owner
+	 * rights - it'd be worse for security to require every user performing
+	 * data loads to be made a table owner - besides being more complex to set
+	 * up.
+	 */
+	EnsureTablePermissions(distributedRelationId, ACL_INSERT);
+
+	/* finally insert placement */
+	InsertShardPlacementRow(shardId, shardState, shardLength, nodeName, nodePort);
+
+	heap_close(relation, NoLock);
+	heap_close(pgDistShard, NoLock);
+
+	PG_RETURN_VOID();
+}
