@@ -40,10 +40,10 @@
 
 
 /* Local functions forward declarations */
-static bool WorkerCreateShard(char *nodeName, uint32 nodePort,
-							  uint64 shardId, List *ddlCommandList);
+static bool WorkerCreateShard(char *nodeName, uint32 nodePort, uint64 shardId,
+							  char *newShardOwner, List *ddlCommandList);
 static bool WorkerShardStats(char *nodeName, uint32 nodePort, Oid relationId,
-							 char *shardName, uint64 *shardLength,
+							 char *shardName, uint64 *shardSize,
 							 text **shardMinValue, text **shardMaxValue);
 static uint64 WorkerTableSize(char *nodeName, uint32 nodePort, Oid relationId,
 							  char *tableName);
@@ -54,6 +54,7 @@ static StringInfo WorkerPartitionValue(char *nodeName, uint32 nodePort, Oid rela
 /* exports for SQL callable functions */
 PG_FUNCTION_INFO_V1(master_create_empty_shard);
 PG_FUNCTION_INFO_V1(master_append_table_to_shard);
+PG_FUNCTION_INFO_V1(master_update_shard_statistics);
 
 
 /*
@@ -82,6 +83,9 @@ master_create_empty_shard(PG_FUNCTION_ARGS)
 	char storageType = SHARD_STORAGE_TABLE;
 
 	Oid relationId = ResolveRelationId(relationNameText);
+	char *relationOwner = TableOwner(relationId);
+
+	EnsureTablePermissions(relationId, ACL_INSERT);
 	CheckDistributedTable(relationId);
 
 	if (CStoreTable(relationId))
@@ -127,8 +131,8 @@ master_create_empty_shard(PG_FUNCTION_ARGS)
 		candidateNodeCount++;
 	}
 
-	CreateShardPlacements(shardId, ddlEventList, candidateNodeList, 0,
-						  ShardReplicationFactor);
+	CreateShardPlacements(shardId, ddlEventList, relationOwner,
+						  candidateNodeList, 0, ShardReplicationFactor);
 
 	InsertShardRow(relationId, shardId, storageType, nullMinValue, nullMaxValue);
 
@@ -159,14 +163,10 @@ master_append_table_to_shard(PG_FUNCTION_ARGS)
 	List *succeededPlacementList = NIL;
 	List *failedPlacementList = NIL;
 	ListCell *shardPlacementCell = NULL;
-	ListCell *succeededPlacementCell = NULL;
 	ListCell *failedPlacementCell = NULL;
-	bool statsOK = false;
-	uint64 newShardLength = 0;
+	uint64 newShardSize = 0;
 	uint64 shardMaxSizeInBytes = 0;
 	float4 shardFillLevel = 0.0;
-	text *newMinValue = NULL;
-	text *newMaxValue = NULL;
 	char partitionMethod = 0;
 
 	ShardInterval *shardInterval = LoadShardInterval(shardId);
@@ -174,6 +174,9 @@ master_append_table_to_shard(PG_FUNCTION_ARGS)
 	bool cstoreTable = CStoreTable(relationId);
 
 	char storageType = shardInterval->storageType;
+
+	EnsureTablePermissions(relationId, ACL_INSERT);
+
 	if (storageType != SHARD_STORAGE_TABLE && !cstoreTable)
 	{
 		ereport(ERROR, (errmsg("cannot append to shardId " UINT64_FORMAT, shardId),
@@ -225,7 +228,9 @@ master_append_table_to_shard(PG_FUNCTION_ARGS)
 						 quote_literal_cstr(sourceTableName),
 						 quote_literal_cstr(sourceNodeName), sourceNodePort);
 
-		queryResultList = ExecuteRemoteQuery(workerName, workerPort, workerAppendQuery);
+		/* inserting data should be performed by the current user */
+		queryResultList = ExecuteRemoteQuery(workerName, workerPort, NULL,
+											 workerAppendQuery);
 		if (queryResultList != NIL)
 		{
 			succeededPlacementList = lappend(succeededPlacementList, shardPlacement);
@@ -264,69 +269,33 @@ master_append_table_to_shard(PG_FUNCTION_ARGS)
 
 	RESUME_INTERRUPTS();
 
-	/* get appended shard's statistics from a shard placement */
-	foreach(succeededPlacementCell, succeededPlacementList)
-	{
-		ShardPlacement *placement = (ShardPlacement *) lfirst(succeededPlacementCell);
-		char *workerName = placement->nodeName;
-		uint32 workerPort = placement->nodePort;
-
-		statsOK = WorkerShardStats(workerName, workerPort, relationId, shardName,
-								   &newShardLength, &newMinValue, &newMaxValue);
-		if (statsOK)
-		{
-			break;
-		}
-	}
-
-	/*
-	 * If for some reason we appended data to a shard, but failed to retrieve
-	 * statistics we just WARN here to avoid losing shard-state updates. Note
-	 * that this means we will return 0 as the shard fill-factor, and this shard
-	 * also won't be pruned as the statistics will be empty. If the failure was
-	 * transient, a subsequent append call will fetch the correct statistics.
-	 */
-	if (!statsOK)
-	{
-		ereport(WARNING, (errmsg("could not get statistics for shard placement"),
-						  errdetail("Setting shard statistics to NULL")));
-	}
-
-	/* make sure we don't process cancel signals */
-	HOLD_INTERRUPTS();
-
-	/* update metadata for each shard placement we appended to */
-	succeededPlacementCell = NULL;
-	foreach(succeededPlacementCell, succeededPlacementList)
-	{
-		ShardPlacement *placement = (ShardPlacement *) lfirst(succeededPlacementCell);
-		char *workerName = placement->nodeName;
-		uint32 workerPort = placement->nodePort;
-
-		DeleteShardPlacementRow(shardId, workerName, workerPort);
-		InsertShardPlacementRow(shardId, FILE_FINALIZED, newShardLength,
-								workerName, workerPort);
-	}
-
-	DeleteShardRow(shardId);
-	InsertShardRow(relationId, shardId, storageType, newMinValue, newMaxValue);
-
-	if (QueryCancelPending)
-	{
-		ereport(WARNING, (errmsg("cancel requests are ignored during table appends")));
-		QueryCancelPending = false;
-	}
-
-	RESUME_INTERRUPTS();
+	/* update shard statistics and get new shard size */
+	newShardSize = UpdateShardStatistics(shardId);
 
 	/* calculate ratio of current shard size compared to shard max size */
 	shardMaxSizeInBytes = (int64) ShardMaxSize * 1024L;
-	shardFillLevel = ((float4) newShardLength / (float4) shardMaxSizeInBytes);
+	shardFillLevel = ((float4) newShardSize / (float4) shardMaxSizeInBytes);
 
 	PG_RETURN_FLOAT4(shardFillLevel);
 }
 
 
+/*
+ * master_update_shard_statistics updates metadata (shard size and shard min/max
+ * values) of the given shard and returns the updated shard size.
+ */
+Datum
+master_update_shard_statistics(PG_FUNCTION_ARGS)
+{
+	int64 shardId = PG_GETARG_INT64(0);
+	uint64 shardSize = 0;
+
+	shardSize = UpdateShardStatistics(shardId);
+
+	PG_RETURN_INT64(shardSize);
+}
+
+
 /*
  * CheckDistributedTable checks if the given relationId corresponds to a
  * distributed table. If it does not, the function errors out.
@@ -361,8 +330,8 @@ CheckDistributedTable(Oid relationId)
  * nodes if some DDL commands had been successful).
  */
 void
-CreateShardPlacements(int64 shardId, List *ddlEventList, List *workerNodeList,
-					  int workerStartIndex, int replicationFactor)
+CreateShardPlacements(int64 shardId, List *ddlEventList, char *newPlacementOwner,
+					  List *workerNodeList, int workerStartIndex, int replicationFactor)
 {
 	int attemptCount = replicationFactor;
 	int workerNodeCount = list_length(workerNodeList);
@@ -382,7 +351,8 @@ CreateShardPlacements(int64 shardId, List *ddlEventList, List *workerNodeList,
 		char *nodeName = workerNode->workerName;
 		uint32 nodePort = workerNode->workerPort;
 
-		bool created = WorkerCreateShard(nodeName, nodePort, shardId, ddlEventList);
+		bool created = WorkerCreateShard(nodeName, nodePort, shardId, newPlacementOwner,
+										 ddlEventList);
 		if (created)
 		{
 			const RelayFileState shardState = FILE_FINALIZED;
@@ -418,8 +388,8 @@ CreateShardPlacements(int64 shardId, List *ddlEventList, List *workerNodeList,
  * each DDL command, and could leave the shard in an half-initialized state.
  */
 static bool
-WorkerCreateShard(char *nodeName, uint32 nodePort,
-				  uint64 shardId, List *ddlCommandList)
+WorkerCreateShard(char *nodeName, uint32 nodePort, uint64 shardId,
+				  char *newShardOwner, List *ddlCommandList)
 {
 	bool shardCreated = true;
 	ListCell *ddlCommandCell = NULL;
@@ -434,7 +404,8 @@ WorkerCreateShard(char *nodeName, uint32 nodePort,
 		appendStringInfo(applyDDLCommand, WORKER_APPLY_SHARD_DDL_COMMAND,
 						 shardId, escapedDDLCommand);
 
-		queryResultList = ExecuteRemoteQuery(nodeName, nodePort, applyDDLCommand);
+		queryResultList = ExecuteRemoteQuery(nodeName, nodePort, newShardOwner,
+											 applyDDLCommand);
 		if (queryResultList == NIL)
 		{
 			shardCreated = false;
@@ -446,13 +417,107 @@ WorkerCreateShard(char *nodeName, uint32 nodePort,
 }
 
 
+/*
+ * UpdateShardStatistics updates metadata (shard size and shard min/max values)
+ * of the given shard and returns the updated shard size.
+ */
+uint64
+UpdateShardStatistics(int64 shardId)
+{
+	ShardInterval *shardInterval = LoadShardInterval(shardId);
+	Oid relationId = shardInterval->relationId;
+	char storageType = shardInterval->storageType;
+	char *shardQualifiedName = NULL;
+	List *shardPlacementList = NIL;
+	ListCell *shardPlacementCell = NULL;
+	bool statsOK = false;
+	uint64 shardSize = 0;
+	text *minValue = NULL;
+	text *maxValue = NULL;
+
+	/* if shard doesn't have an alias, extend regular table name */
+	shardQualifiedName = LoadShardAlias(relationId, shardId);
+	if (shardQualifiedName == NULL)
+	{
+		char *relationName = get_rel_name(relationId);
+
+		Oid schemaId = get_rel_namespace(relationId);
+		char *schemaName = get_namespace_name(schemaId);
+
+		shardQualifiedName = quote_qualified_identifier(schemaName, relationName);
+
+		AppendShardIdToName(&shardQualifiedName, shardId);
+	}
+
+	shardPlacementList = FinalizedShardPlacementList(shardId);
+
+	/* get shard's statistics from a shard placement */
+	foreach(shardPlacementCell, shardPlacementList)
+	{
+		ShardPlacement *placement = (ShardPlacement *) lfirst(shardPlacementCell);
+		char *workerName = placement->nodeName;
+		uint32 workerPort = placement->nodePort;
+
+		statsOK = WorkerShardStats(workerName, workerPort, relationId, shardQualifiedName,
+								   &shardSize, &minValue, &maxValue);
+		if (statsOK)
+		{
+			break;
+		}
+	}
+
+	/*
+	 * If for some reason we appended data to a shard, but failed to retrieve
+	 * statistics we just WARN here to avoid losing shard-state updates. Note
+	 * that this means we will return 0 as the shard fill-factor, and this shard
+	 * also won't be pruned as the statistics will be empty. If the failure was
+	 * transient, a subsequent append call will fetch the correct statistics.
+	 */
+	if (!statsOK)
+	{
+		ereport(WARNING, (errmsg("could not get statistics for shard %s",
+								 shardQualifiedName),
+						  errdetail("Setting shard statistics to NULL")));
+	}
+
+	/* make sure we don't process cancel signals */
+	HOLD_INTERRUPTS();
+
+	/* update metadata for each shard placement we appended to */
+	shardPlacementCell = NULL;
+	foreach(shardPlacementCell, shardPlacementList)
+	{
+		ShardPlacement *placement = (ShardPlacement *) lfirst(shardPlacementCell);
+		char *workerName = placement->nodeName;
+		uint32 workerPort = placement->nodePort;
+
+		DeleteShardPlacementRow(shardId, workerName, workerPort);
+		InsertShardPlacementRow(shardId, FILE_FINALIZED, shardSize,
+								workerName, workerPort);
+	}
+
+	DeleteShardRow(shardId);
+	InsertShardRow(relationId, shardId, storageType, minValue, maxValue);
+
+	if (QueryCancelPending)
+	{
+		ereport(WARNING, (errmsg("cancel requests are ignored during metadata update")));
+		QueryCancelPending = false;
+	}
+
+	RESUME_INTERRUPTS();
+
+	return shardSize;
+}
+
+
 /*
  * WorkerShardStats queries the worker node, and retrieves shard statistics that
  * we assume have changed after new table data have been appended to the shard.
  */
 static bool
 WorkerShardStats(char *nodeName, uint32 nodePort, Oid relationId, char *shardName,
-				 uint64 *shardLength, text **shardMinValue, text **shardMaxValue)
+				 uint64 *shardSize, text **shardMinValue, text **shardMaxValue)
 {
 	bool shardStatsOK = true;
 
@@ -464,7 +529,7 @@ WorkerShardStats(char *nodeName, uint32 nodePort, Oid relationId, char *shardNam
 		StringInfo maxValue = WorkerPartitionValue(nodeName, nodePort, relationId,
 												   shardName, SHARD_MAX_VALUE_QUERY);
 
-		(*shardLength) = tableSize;
+		(*shardSize) = tableSize;
 		(*shardMinValue) = cstring_to_text_with_len(minValue->data, minValue->len);
 		(*shardMaxValue) = cstring_to_text_with_len(maxValue->data, maxValue->len);
 	}
@@ -502,7 +567,7 @@ WorkerTableSize(char *nodeName, uint32 nodePort, Oid relationId, char *tableName
 		appendStringInfo(tableSizeQuery, SHARD_TABLE_SIZE_QUERY, tableName);
 	}
 
-	queryResultList = ExecuteRemoteQuery(nodeName, nodePort, tableSizeQuery);
+	queryResultList = ExecuteRemoteQuery(nodeName, nodePort, NULL, tableSizeQuery);
 	if (queryResultList == NIL)
 	{
 		ereport(ERROR, (errmsg("could not receive table size from node "
@@ -548,7 +613,7 @@ WorkerPartitionValue(char *nodeName, uint32 nodePort, Oid relationId,
 	 * simply casts the results to a (char *). If the user partitioned the table
 	 * on a binary byte array, this approach fails and should be fixed.
 	 */
-	queryResultList = ExecuteRemoteQuery(nodeName, nodePort, partitionValueQuery);
+	queryResultList = ExecuteRemoteQuery(nodeName, nodePort, NULL, partitionValueQuery);
 	if (queryResultList == NIL)
 	{
 		ereport(ERROR, (errmsg("could not receive shard min/max values from node "
