@@ -56,7 +56,6 @@ static List * MergeShardIntervals(List *leftShardIntervalList,
 								  List *rightShardIntervalList, JoinType joinType);
 static bool ShardIntervalsMatch(List *leftShardIntervalList,
 								List *rightShardIntervalList);
-static List * LoadSortedShardIntervalList(Oid relationId);
 static List * JoinOrderForTable(TableEntry *firstTable, List *tableEntryList,
 								List *joinClauseList);
 static List * BestJoinOrder(List *candidateJoinOrders);
@@ -123,6 +122,22 @@ FixedJoinOrderList(FromExpr *fromExpr, List *tableEntryList)
 	List *joinedTableList = NIL;
 	JoinOrderNode *firstJoinNode = NULL;
 	JoinOrderNode *currentJoinNode = NULL;
+	ListCell *tableEntryCell = NULL;
+
+	foreach(tableEntryCell, tableEntryList)
+	{
+		TableEntry *rangeTableEntry = (TableEntry *) lfirst(tableEntryCell);
+		Oid relationId = rangeTableEntry->relationId;
+		DistTableCacheEntry *cacheEntry = DistributedTableCacheEntry(relationId);
+
+		if (cacheEntry->hasUninitializedShardInterval)
+		{
+			ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+							errmsg("cannot perform distributed planning on this query"),
+							errdetail("Shards of relations in outer join queries must "
+									  "have shard min/max values.")));
+		}
+	}
 
 	/* get the FROM section as a flattened list of JoinExpr nodes */
 	joinList = JoinExprList(fromExpr);
@@ -159,8 +174,8 @@ FixedJoinOrderList(FromExpr *fromExpr, List *tableEntryList)
 			joinClauseList = list_concat(joinClauseList, joinWhereClauseList);
 		}
 
-		/* get the list of shards to check broadcast/local join possibility */
-		candidateShardList = LoadSortedShardIntervalList(nextTable->relationId);
+		/* get the sorted list of shards to check broadcast/local join possibility */
+		candidateShardList = LoadShardIntervalList(nextTable->relationId);
 
 		/* find the best join rule type */
 		nextJoinNode = EvaluateJoinRules(joinedTableList, currentJoinNode,
@@ -268,8 +283,7 @@ CreateFirstJoinOrderNode(FromExpr *fromExpr, List *tableEntryList)
 									  firstPartitionColumn,
 									  firstPartitionMethod);
 
-	firstJoinNode->shardIntervalList =
-		LoadSortedShardIntervalList(firstTable->relationId);
+	firstJoinNode->shardIntervalList = LoadShardIntervalList(firstTable->relationId);
 
 	return firstJoinNode;
 }
@@ -462,40 +476,6 @@ MergeShardIntervals(List *leftShardIntervalList, List *rightShardIntervalList,
 }
 
 
-/*
- * LoadSortedShardIntervalList loads a list of shard intervals from the metadata
- * and sorts the list by the minimum value of the intervals.
- */
-static List *
-LoadSortedShardIntervalList(Oid relationId)
-{
-	List *shardIntervalList = NIL;
-	int shardCount = 0;
-	int intervalIndex = 0;
-	ShardInterval **sortedShardIntervalArray = NULL;
-	List *sortedShardIntervalList = NIL;
-
-	shardIntervalList = LoadShardIntervalList(relationId);
-
-	shardCount = list_length(shardIntervalList);
-	if (shardCount <= 1)
-	{
-		return shardIntervalList;
-	}
-
-	sortedShardIntervalArray = SortedShardIntervalArray(shardIntervalList);
-
-	for (intervalIndex = 0; intervalIndex < shardCount; intervalIndex++)
-	{
-		ShardInterval *shardInterval = sortedShardIntervalArray[intervalIndex];
-
-		sortedShardIntervalList = lappend(sortedShardIntervalList, shardInterval);
-	}
-
-	return sortedShardIntervalList;
-}
-
-
 /*
  * JoinOnColumns determines whether two columns are joined by a given join clause
  * list.
@@ -750,7 +730,7 @@ BestJoinOrder(List *candidateJoinOrders)
 	List *bestJoinOrder = NULL;
 	uint32 ruleTypeIndex = 0;
 	uint32 highestValidIndex = JOIN_RULE_LAST - 1;
-	uint32 candidateCount = 0;
+	uint32 candidateCount PG_USED_FOR_ASSERTS_ONLY = 0;
 
 	/*
 	 * We start with the highest ranking rule type (cartesian product), and walk
