@@ -19,14 +19,16 @@
 #include "commands/explain.h"
 #include "executor/executor.h"
 #include "distributed/master_protocol.h"
-#include "distributed/modify_planner.h"
+#include "distributed/multi_copy.h"
 #include "distributed/multi_executor.h"
 #include "distributed/multi_explain.h"
 #include "distributed/multi_join_order.h"
 #include "distributed/multi_logical_optimizer.h"
 #include "distributed/multi_planner.h"
-#include  "distributed/multi_router_executor.h"
+#include "distributed/multi_router_executor.h"
+#include "distributed/multi_router_planner.h"
 #include "distributed/multi_server_executor.h"
+#include "distributed/multi_transaction.h"
 #include "distributed/multi_utility.h"
 #include "distributed/task_tracker.h"
 #include "distributed/worker_manager.h"
@@ -57,7 +59,6 @@ static const struct config_enum_entry task_assignment_policy_options[] = {
 static const struct config_enum_entry task_executor_type_options[] = {
 	{ "real-time", MULTI_EXECUTOR_REAL_TIME, false },
 	{ "task-tracker", MULTI_EXECUTOR_TASK_TRACKER, false },
-	{ "router", MULTI_EXECUTOR_ROUTER, false },
 	{ NULL, 0, false }
 };
 
@@ -67,6 +68,12 @@ static const struct config_enum_entry shard_placement_policy_options[] = {
 	{ NULL, 0, false }
 };
 
+static const struct config_enum_entry multi_shard_commit_protocol_options[] = {
+	{ "1pc", COMMIT_PROTOCOL_1PC, false },
+	{ "2pc", COMMIT_PROTOCOL_2PC, false },
+	{ NULL, 0, false }
+};
+
 
 /* shared library initialization function */
 void
@@ -75,7 +82,10 @@ _PG_init(void)
 	if (!process_shared_preload_libraries_in_progress)
 	{
 		ereport(ERROR, (errmsg("Citus can only be loaded via shared_preload_libraries"),
-						errhint("Add citus to shared_preload_libraries.")));
+						errhint("Add citus to shared_preload_libraries configuration "
+								"variable in postgresql.conf in master and workers. Note "
+								"that citus should be at the beginning of "
+								"shared_preload_libraries.")));
 	}
 
 	/*
@@ -262,6 +272,31 @@ RegisterCitusConfigVariables(void)
 		GUC_NO_SHOW_ALL,
 		NULL, NULL, NULL);
 
+	DefineCustomBoolVariable(
+		"citus.explain_distributed_queries",
+		gettext_noop("Enables Explain for distributed queries."),
+		gettext_noop("When enabled, the Explain command shows remote and local "
+					 "plans when used with a distributed query. It is enabled "
+					 "by default, but can be disabled for regression tests."),
+		&ExplainDistributedQueries,
+		true,
+		PGC_USERSET,
+		GUC_NO_SHOW_ALL,
+		NULL, NULL, NULL);
+
+	DefineCustomBoolVariable(
+		"citus.explain_all_tasks",
+		gettext_noop("Enables showing output for all tasks in Explain."),
+		gettext_noop("The Explain command for distributed queries shows "
+					 "the remote plan for a single task by default. When "
+					 "this configuration entry is enabled, the plan for "
+					 "all tasks is shown, but the Explain takes longer."),
+		&ExplainAllTasks,
+		false,
+		PGC_USERSET,
+		0,
+		NULL, NULL, NULL);
+
 	DefineCustomBoolVariable(
 		"citus.all_modifications_commutative",
 		gettext_noop("Bypasses commutativity checks when enabled"),
@@ -437,6 +472,22 @@ RegisterCitusConfigVariables(void)
 		0,
 		NULL, NULL, NULL);
 
+	DefineCustomEnumVariable(
+		"citus.multi_shard_commit_protocol",
+		gettext_noop("Sets the commit protocol for commands modifying multiple shards."),
+		gettext_noop("When a failure occurs during commands that modify multiple "
+					 "shards (currently, only COPY on distributed tables modifies more "
+					 "than one shard), two-phase commit is required to ensure data is "
+					 "never lost. Change this setting to '2pc' from its default '1pc' to "
+					 "enable 2 PC. You must also set max_prepared_transactions on the "
+					 "worker nodes. Recovery from failed 2PCs is currently manual."),
+		&MultiShardCommitProtocol,
+		COMMIT_PROTOCOL_1PC,
+		multi_shard_commit_protocol_options,
+		PGC_USERSET,
+		0,
+		NULL, NULL, NULL);
+
 	DefineCustomEnumVariable(
 		"citus.task_assignment_policy",
 		gettext_noop("Sets the policy to use when assigning tasks to worker nodes."),
@@ -457,14 +508,13 @@ RegisterCitusConfigVariables(void)
 	DefineCustomEnumVariable(
 		"citus.task_executor_type",
 		gettext_noop("Sets the executor type to be used for distributed queries."),
-		gettext_noop("The master node chooses between three different executor types "
-					 "when executing a distributed query. The router executor is "
-					 "optimal for simple key-value lookups on a single shard. The "
-					 "real-time executor is optimal for queries that involve "
-					 "aggregations and/or co-located joins on multiple shards. The "
-					 "task-tracker executor is optimal for long-running, complex "
-					 "queries that touch thousands of shards and/or that involve "
-					 "table repartitioning."),
+		gettext_noop("The master node chooses between two different executor types "
+					 "when executing a distributed query.The real-time executor is "
+					 "optimal for simple key-value lookup queries and queries that "
+					 "involve aggregations and/or co-located joins on multiple shards. "
+					 "The task-tracker executor is optimal for long-running, complex "
+					 "queries that touch thousands of shards and/or that involve table "
+					 "repartitioning."),
 		&TaskExecutorType,
 		MULTI_EXECUTOR_REAL_TIME,
 		task_executor_type_options,
