@@ -9,10 +9,16 @@
 
 #include "postgres.h"
 
+#include "miscadmin.h"
+
 #include "access/genam.h"
 #include "access/heapam.h"
 #include "access/htup_details.h"
+#include "access/nbtree.h"
+#include "access/xact.h"
+#include "access/sysattr.h"
 #include "catalog/indexing.h"
+#include "catalog/pg_extension.h"
 #include "catalog/pg_namespace.h"
 #include "catalog/pg_type.h"
 #include "commands/extension.h"
@@ -21,6 +27,8 @@
 #include "distributed/metadata_cache.h"
 #include "distributed/pg_dist_partition.h"
 #include "distributed/pg_dist_shard.h"
+#include "distributed/shardinterval_utils.h"
+#include "distributed/worker_protocol.h"
 #include "parser/parse_func.h"
 #include "utils/builtins.h"
 #include "utils/catcache.h"
@@ -33,8 +41,20 @@
 #include "utils/relfilenodemap.h"
 #include "utils/relmapper.h"
 #include "utils/syscache.h"
+#include "utils/typcache.h"
 
 
+/* state which should be cleared upon DROP EXTENSION */
+static bool extensionLoaded = false;
+static Oid distShardRelationId = InvalidOid;
+static Oid distShardPlacementRelationId = InvalidOid;
+static Oid distPartitionRelationId = InvalidOid;
+static Oid distPartitionLogicalRelidIndexId = InvalidOid;
+static Oid distShardLogicalRelidIndexId = InvalidOid;
+static Oid distShardShardidIndexId = InvalidOid;
+static Oid distShardPlacementShardidIndexId = InvalidOid;
+static Oid extraDataContainerFuncId = InvalidOid;
+
 /* Hash table for informations about each partition */
 static HTAB *DistTableCacheHash = NULL;
 
@@ -45,6 +65,16 @@ static ScanKeyData DistShardScanKey[1];
 
 /* local function forward declarations */
 static DistTableCacheEntry * LookupDistTableCacheEntry(Oid relationId);
+static FmgrInfo * ShardIntervalCompareFunction(ShardInterval **shardIntervalArray,
+											   char partitionMethod);
+static ShardInterval ** SortShardIntervalArray(ShardInterval **shardIntervalArray,
+											   int shardCount,
+											   FmgrInfo *
+											   shardIntervalSortCompareFunction);
+static bool HasUniformHashDistribution(ShardInterval **shardIntervalArray,
+									   int shardIntervalArrayLength);
+static bool HasUninitializedShardInterval(ShardInterval **sortedShardIntervalArray,
+										  int shardCount);
 static void InitializeDistTableCache(void);
 static void ResetDistTableCacheEntry(DistTableCacheEntry *cacheEntry);
 static void InvalidateDistRelationCacheCallback(Datum argument, Oid relationId);
@@ -190,7 +220,12 @@ LookupDistTableCacheEntry(Oid relationId)
 	char partitionMethod = 0;
 	List *distShardTupleList = NIL;
 	int shardIntervalArrayLength = 0;
-	ShardInterval *shardIntervalArray = NULL;
+	ShardInterval **shardIntervalArray = NULL;
+	ShardInterval **sortedShardIntervalArray = NULL;
+	FmgrInfo *shardIntervalCompareFunction = NULL;
+	FmgrInfo *hashFunction = NULL;
+	bool hasUninitializedShardInterval = false;
+	bool hasUniformHashDistribution = false;
 	void *hashKey = (void *) &relationId;
 
 	if (DistTableCacheHash == NULL)
@@ -245,7 +280,7 @@ LookupDistTableCacheEntry(Oid relationId)
 
 		shardIntervalArray = MemoryContextAllocZero(CacheMemoryContext,
 													shardIntervalArrayLength *
-													sizeof(ShardInterval));
+													sizeof(ShardInterval *));
 
 		foreach(distShardTupleCell, distShardTupleList)
 		{
@@ -254,9 +289,12 @@ LookupDistTableCacheEntry(Oid relationId)
 																distShardTupleDesc,
 																intervalTypeId,
 																intervalTypeMod);
+			ShardInterval *newShardInterval = NULL;
 			MemoryContext oldContext = MemoryContextSwitchTo(CacheMemoryContext);
 
-			CopyShardInterval(shardInterval, &shardIntervalArray[arrayIndex]);
+			newShardInterval = (ShardInterval *) palloc0(sizeof(ShardInterval));
+			CopyShardInterval(shardInterval, newShardInterval);
+			shardIntervalArray[arrayIndex] = newShardInterval;
 
 			MemoryContextSwitchTo(oldContext);
 
@@ -268,6 +306,50 @@ LookupDistTableCacheEntry(Oid relationId)
 		heap_close(distShardRelation, AccessShareLock);
 	}
 
+	/* decide and allocate interval comparison function */
+	if (shardIntervalArrayLength > 0)
+	{
+		MemoryContext oldContext = CurrentMemoryContext;
+
+		/* allocate the comparison function in the cache context */
+		oldContext = MemoryContextSwitchTo(CacheMemoryContext);
+
+		shardIntervalCompareFunction = ShardIntervalCompareFunction(shardIntervalArray,
+																	partitionMethod);
+
+		MemoryContextSwitchTo(oldContext);
+	}
+
+	/* sort the interval array */
+	sortedShardIntervalArray = SortShardIntervalArray(shardIntervalArray,
+													  shardIntervalArrayLength,
+													  shardIntervalCompareFunction);
+
+	/* check if there exists any shard intervals with no min/max values */
+	hasUninitializedShardInterval =
+		HasUninitializedShardInterval(sortedShardIntervalArray, shardIntervalArrayLength);
+
+	/* we only need hash functions for hash distributed tables */
+	if (partitionMethod == DISTRIBUTE_BY_HASH)
+	{
+		TypeCacheEntry *typeEntry = NULL;
+		Node *partitionNode = stringToNode(partitionKeyString);
+		Var *partitionColumn = (Var *) partitionNode;
+		Assert(IsA(partitionNode, Var));
+		typeEntry = lookup_type_cache(partitionColumn->vartype,
+									  TYPECACHE_HASH_PROC_FINFO);
+
+		hashFunction = MemoryContextAllocZero(CacheMemoryContext,
+											  sizeof(FmgrInfo));
+
+		fmgr_info_copy(hashFunction, &(typeEntry->hash_proc_finfo), CacheMemoryContext);
+
+		/* check the shard distribution for hash partitioned tables */
+		hasUniformHashDistribution =
+			HasUniformHashDistribution(sortedShardIntervalArray,
+									   shardIntervalArrayLength);
+	}
+
 	cacheEntry = hash_search(DistTableCacheHash, hashKey, HASH_ENTER, NULL);
 
 	/* zero out entry, but not the key part */
@@ -286,27 +368,159 @@ LookupDistTableCacheEntry(Oid relationId)
 		cacheEntry->partitionKeyString = partitionKeyString;
 		cacheEntry->partitionMethod = partitionMethod;
 		cacheEntry->shardIntervalArrayLength = shardIntervalArrayLength;
-		cacheEntry->shardIntervalArray = shardIntervalArray;
+		cacheEntry->sortedShardIntervalArray = sortedShardIntervalArray;
+		cacheEntry->shardIntervalCompareFunction = shardIntervalCompareFunction;
+		cacheEntry->hashFunction = hashFunction;
+		cacheEntry->hasUninitializedShardInterval = hasUninitializedShardInterval;
+		cacheEntry->hasUniformHashDistribution = hasUniformHashDistribution;
 	}
 
 	return cacheEntry;
 }
 
 
+/*
+ * ShardIntervalCompareFunction returns the appropriate compare function for the
+ * partition column type. In case of hash-partitioning, it always returns the compare
+ * function for integers. Callers of this function has to ensure that shardIntervalArray
+ * has at least one element.
+ */
+static FmgrInfo *
+ShardIntervalCompareFunction(ShardInterval **shardIntervalArray, char partitionMethod)
+{
+	FmgrInfo *shardIntervalCompareFunction = NULL;
+	Oid comparisonTypeId = InvalidOid;
+
+	Assert(shardIntervalArray != NULL);
+
+	if (partitionMethod == DISTRIBUTE_BY_HASH)
+	{
+		comparisonTypeId = INT4OID;
+	}
+	else
+	{
+		ShardInterval *shardInterval = shardIntervalArray[0];
+		comparisonTypeId = shardInterval->valueTypeId;
+	}
+
+	shardIntervalCompareFunction = GetFunctionInfo(comparisonTypeId, BTREE_AM_OID,
+												   BTORDER_PROC);
+
+	return shardIntervalCompareFunction;
+}
+
+
+/*
+ * SortedShardIntervalArray sorts the input shardIntervalArray. Shard intervals with
+ * no min/max values are placed at the end of the array.
+ */
+static ShardInterval **
+SortShardIntervalArray(ShardInterval **shardIntervalArray, int shardCount,
+					   FmgrInfo *shardIntervalSortCompareFunction)
+{
+	ShardInterval **sortedShardIntervalArray = NULL;
+
+	/* short cut if there are no shard intervals in the array */
+	if (shardCount == 0)
+	{
+		return shardIntervalArray;
+	}
+
+	/* if a shard doesn't have min/max values, it's placed in the end of the array */
+	qsort_arg(shardIntervalArray, shardCount, sizeof(ShardInterval *),
+			  (qsort_arg_comparator) CompareShardIntervals,
+			  (void *) shardIntervalSortCompareFunction);
+
+	sortedShardIntervalArray = shardIntervalArray;
+
+	return sortedShardIntervalArray;
+}
+
+
+/*
+ * HasUniformHashDistribution determines whether the given list of sorted shards
+ * has a uniform hash distribution, as produced by master_create_worker_shards for
+ * hash partitioned tables.
+ */
+static bool
+HasUniformHashDistribution(ShardInterval **shardIntervalArray,
+						   int shardIntervalArrayLength)
+{
+	uint64 hashTokenIncrement = 0;
+	int shardIndex = 0;
+
+	/* if there are no shards, there is no uniform distribution */
+	if (shardIntervalArrayLength == 0)
+	{
+		return false;
+	}
+
+	/* calculate the hash token increment */
+	hashTokenIncrement = HASH_TOKEN_COUNT / shardIntervalArrayLength;
+
+	for (shardIndex = 0; shardIndex < shardIntervalArrayLength; shardIndex++)
+	{
+		ShardInterval *shardInterval = shardIntervalArray[shardIndex];
+		int32 shardMinHashToken = INT32_MIN + (shardIndex * hashTokenIncrement);
+		int32 shardMaxHashToken = shardMinHashToken + (hashTokenIncrement - 1);
+
+		if (shardIndex == (shardIntervalArrayLength - 1))
+		{
+			shardMaxHashToken = INT32_MAX;
+		}
+
+		if (DatumGetInt32(shardInterval->minValue) != shardMinHashToken ||
+			DatumGetInt32(shardInterval->maxValue) != shardMaxHashToken)
+		{
+			return false;
+		}
+	}
+
+	return true;
+}
+
+
+/*
+ * HasUninitializedShardInterval returns true if all the elements of the
+ * sortedShardIntervalArray has min/max values. Callers of the function must
+ * ensure that input shard interval array is sorted on shardminvalue and uninitialized
+ * shard intervals are at the end of the array.
+ */
+static bool
+HasUninitializedShardInterval(ShardInterval **sortedShardIntervalArray, int shardCount)
+{
+	bool hasUninitializedShardInterval = false;
+	ShardInterval *lastShardInterval = NULL;
+
+	if (shardCount == 0)
+	{
+		return hasUninitializedShardInterval;
+	}
+
+	Assert(sortedShardIntervalArray != NULL);
+
+	/*
+	 * Since the shard interval array is sorted, and uninitialized ones stored
+	 * in the end of the array, checking the last element is enough.
+	 */
+	lastShardInterval = sortedShardIntervalArray[shardCount - 1];
+	if (!lastShardInterval->minValueExists || !lastShardInterval->maxValueExists)
+	{
+		hasUninitializedShardInterval = true;
+	}
+
+	return hasUninitializedShardInterval;
+}
+
+
 /*
  * CitusHasBeenLoaded returns true if the citus extension has been created
  * in the current database and the extension script has been executed. Otherwise,
  * it returns false. The result is cached as this is called very frequently.
- *
- * NB: The way this is cached means the result will be wrong after the
- * extension is dropped. A reconnect fixes that though, so that seems
- * acceptable.
  */
 bool
 CitusHasBeenLoaded(void)
 {
-	static bool extensionLoaded = false;
-
 	/* recheck presence until citus has been loaded */
 	if (!extensionLoaded)
 	{
@@ -329,6 +543,19 @@ CitusHasBeenLoaded(void)
 		}
 
 		extensionLoaded = extensionPresent && extensionScriptExecuted;
+
+		if (extensionLoaded)
+		{
+			/*
+			 * InvalidateDistRelationCacheCallback resets state such as extensionLoaded
+			 * when it notices changes to pg_dist_partition (which usually indicate
+			 * `DROP EXTENSION citus;` has been run)
+			 *
+			 * Ensure InvalidateDistRelationCacheCallback will notice those changes
+			 * by caching pg_dist_partition's oid.
+			 */
+			DistPartitionRelationId();
+		}
 	}
 
 	return extensionLoaded;
@@ -339,11 +566,9 @@ CitusHasBeenLoaded(void)
 Oid
 DistShardRelationId(void)
 {
-	static Oid cachedOid = InvalidOid;
+	CachedRelationLookup("pg_dist_shard", &distShardRelationId);
 
-	CachedRelationLookup("pg_dist_shard", &cachedOid);
-
-	return cachedOid;
+	return distShardRelationId;
 }
 
 
@@ -351,11 +576,9 @@ DistShardRelationId(void)
 Oid
 DistShardPlacementRelationId(void)
 {
-	static Oid cachedOid = InvalidOid;
+	CachedRelationLookup("pg_dist_shard_placement", &distShardPlacementRelationId);
 
-	CachedRelationLookup("pg_dist_shard_placement", &cachedOid);
-
-	return cachedOid;
+	return distShardPlacementRelationId;
 }
 
 
@@ -363,11 +586,9 @@ DistShardPlacementRelationId(void)
 Oid
 DistPartitionRelationId(void)
 {
-	static Oid cachedOid = InvalidOid;
+	CachedRelationLookup("pg_dist_partition", &distPartitionRelationId);
 
-	CachedRelationLookup("pg_dist_partition", &cachedOid);
-
-	return cachedOid;
+	return distPartitionRelationId;
 }
 
 
@@ -375,11 +596,10 @@ DistPartitionRelationId(void)
 Oid
 DistPartitionLogicalRelidIndexId(void)
 {
-	static Oid cachedOid = InvalidOid;
+	CachedRelationLookup("pg_dist_partition_logical_relid_index",
+						 &distPartitionLogicalRelidIndexId);
 
-	CachedRelationLookup("pg_dist_partition_logical_relid_index", &cachedOid);
-
-	return cachedOid;
+	return distPartitionLogicalRelidIndexId;
 }
 
 
@@ -387,11 +607,10 @@ DistPartitionLogicalRelidIndexId(void)
 Oid
 DistShardLogicalRelidIndexId(void)
 {
-	static Oid cachedOid = InvalidOid;
+	CachedRelationLookup("pg_dist_shard_logical_relid_index",
+						 &distShardLogicalRelidIndexId);
 
-	CachedRelationLookup("pg_dist_shard_logical_relid_index", &cachedOid);
-
-	return cachedOid;
+	return distShardLogicalRelidIndexId;
 }
 
 
@@ -399,11 +618,9 @@ DistShardLogicalRelidIndexId(void)
 Oid
 DistShardShardidIndexId(void)
 {
-	static Oid cachedOid = InvalidOid;
+	CachedRelationLookup("pg_dist_shard_shardid_index", &distShardShardidIndexId);
 
-	CachedRelationLookup("pg_dist_shard_shardid_index", &cachedOid);
-
-	return cachedOid;
+	return distShardShardidIndexId;
 }
 
 
@@ -411,11 +628,10 @@ DistShardShardidIndexId(void)
 Oid
 DistShardPlacementShardidIndexId(void)
 {
-	static Oid cachedOid = InvalidOid;
+	CachedRelationLookup("pg_dist_shard_placement_shardid_index",
+						 &distShardPlacementShardidIndexId);
 
-	CachedRelationLookup("pg_dist_shard_placement_shardid_index", &cachedOid);
-
-	return cachedOid;
+	return distShardPlacementShardidIndexId;
 }
 
 
@@ -423,18 +639,97 @@ DistShardPlacementShardidIndexId(void)
 Oid
 CitusExtraDataContainerFuncId(void)
 {
-	static Oid cachedOid = 0;
 	List *nameList = NIL;
 	Oid paramOids[1] = { INTERNALOID };
 
-	if (cachedOid == InvalidOid)
+	if (extraDataContainerFuncId == InvalidOid)
 	{
 		nameList = list_make2(makeString("pg_catalog"),
 							  makeString("citus_extradata_container"));
-		cachedOid = LookupFuncName(nameList, 1, paramOids, false);
+		extraDataContainerFuncId = LookupFuncName(nameList, 1, paramOids, false);
 	}
 
-	return cachedOid;
+	return extraDataContainerFuncId;
+}
+
+
+/*
+ * CitusExtensionOwner() returns the owner of the 'citus' extension. That user
+ * is, amongst others, used to perform actions a normal user might not be
+ * allowed to perform.
+ */
+extern Oid
+CitusExtensionOwner(void)
+{
+	Relation relation = NULL;
+	SysScanDesc scandesc;
+	ScanKeyData entry[1];
+	HeapTuple extensionTuple = NULL;
+	Form_pg_extension extensionForm = NULL;
+	static Oid extensionOwner = InvalidOid;
+
+	if (extensionOwner != InvalidOid)
+	{
+		return extensionOwner;
+	}
+
+	relation = heap_open(ExtensionRelationId, AccessShareLock);
+
+	ScanKeyInit(&entry[0],
+				Anum_pg_extension_extname,
+				BTEqualStrategyNumber, F_NAMEEQ,
+				CStringGetDatum("citus"));
+
+	scandesc = systable_beginscan(relation, ExtensionNameIndexId, true,
+								  NULL, 1, entry);
+
+	extensionTuple = systable_getnext(scandesc);
+
+	/* We assume that there can be at most one matching tuple */
+	if (HeapTupleIsValid(extensionTuple))
+	{
+		extensionForm = (Form_pg_extension) GETSTRUCT(extensionTuple);
+
+		/*
+		 * For some operations Citus requires superuser permissions; we use
+		 * the extension owner for that. The extension owner is guaranteed to
+		 * be a superuser (otherwise C functions can't be created), but it'd
+		 * be possible to change the owner. So check that this still a
+		 * superuser.
+		 */
+		if (!superuser_arg(extensionForm->extowner))
+		{
+			ereport(ERROR, (errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+							errmsg("citus extension needs to be owned by superuser")));
+		}
+		extensionOwner = extensionForm->extowner;
+		Assert(OidIsValid(extensionOwner));
+	}
+	else
+	{
+		ereport(ERROR, (errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+						errmsg("citus extension not loaded")));
+	}
+
+	systable_endscan(scandesc);
+
+	heap_close(relation, AccessShareLock);
+
+	return extensionOwner;
+}
+
+
+/* return the  username of the currently active role */
+char *
+CurrentUserName(void)
+{
+	Oid userId = GetUserId();
+
+#if (PG_VERSION_NUM < 90500)
+	return GetUserNameFromId(userId);
+#else
+	return GetUserNameFromId(userId, false);
+#endif
 }
 
 
@@ -442,6 +737,9 @@ CitusExtraDataContainerFuncId(void)
  * master_dist_partition_cache_invalidate is a trigger function that performs
  * relcache invalidations when the contents of pg_dist_partition are changed
  * on the SQL level.
+ *
+ * NB: We decided there is little point in checking permissions here, there
+ * are much easier ways to waste CPU than causing cache invalidations.
  */
 Datum
 master_dist_partition_cache_invalidate(PG_FUNCTION_ARGS)
@@ -500,6 +798,9 @@ master_dist_partition_cache_invalidate(PG_FUNCTION_ARGS)
  * master_dist_shard_cache_invalidate is a trigger function that performs
  * relcache invalidations when the contents of pg_dist_shard are changed
  * on the SQL level.
+ *
+ * NB: We decided there is little point in checking permissions here, there
+ * are much easier ways to waste CPU than causing cache invalidations.
  */
 Datum
 master_dist_shard_cache_invalidate(PG_FUNCTION_ARGS)
@@ -621,7 +922,7 @@ ResetDistTableCacheEntry(DistTableCacheEntry *cacheEntry)
 
 		for (i = 0; i < cacheEntry->shardIntervalArrayLength; i++)
 		{
-			ShardInterval *shardInterval = &cacheEntry->shardIntervalArray[i];
+			ShardInterval *shardInterval = cacheEntry->sortedShardIntervalArray[i];
 			bool valueByVal = shardInterval->valueByVal;
 
 			if (!valueByVal)
@@ -636,11 +937,26 @@ ResetDistTableCacheEntry(DistTableCacheEntry *cacheEntry)
 					pfree(DatumGetPointer(shardInterval->maxValue));
 				}
 			}
+
+			pfree(shardInterval);
 		}
 
-		pfree(cacheEntry->shardIntervalArray);
-		cacheEntry->shardIntervalArray = NULL;
+		pfree(cacheEntry->sortedShardIntervalArray);
+		cacheEntry->sortedShardIntervalArray = NULL;
 		cacheEntry->shardIntervalArrayLength = 0;
+
+		cacheEntry->hasUninitializedShardInterval = false;
+		cacheEntry->hasUniformHashDistribution = false;
+
+		pfree(cacheEntry->shardIntervalCompareFunction);
+		cacheEntry->shardIntervalCompareFunction = NULL;
+
+		/* we only allocated hash function for hash distributed tables */
+		if (cacheEntry->partitionMethod == DISTRIBUTE_BY_HASH)
+		{
+			pfree(cacheEntry->hashFunction);
+			cacheEntry->hashFunction = NULL;
+		}
 	}
 }
 
@@ -677,6 +993,24 @@ InvalidateDistRelationCacheCallback(Datum argument, Oid relationId)
 			cacheEntry->isValid = false;
 		}
 	}
+
+	/*
+	 * If pg_dist_partition is being invalidated drop all state
+	 * This happens pretty rarely, but most importantly happens during
+	 * DROP EXTENSION citus;
+	 */
+	if (relationId != InvalidOid && relationId == distPartitionRelationId)
+	{
+		extensionLoaded = false;
+		distShardRelationId = InvalidOid;
+		distShardPlacementRelationId = InvalidOid;
+		distPartitionRelationId = InvalidOid;
+		distPartitionLogicalRelidIndexId = InvalidOid;
+		distShardLogicalRelidIndexId = InvalidOid;
+		distShardShardidIndexId = InvalidOid;
+		distShardPlacementShardidIndexId = InvalidOid;
+		extraDataContainerFuncId = InvalidOid;
+	}
 }
 
 
@@ -878,10 +1212,6 @@ TupleToShardInterval(HeapTuple heapTuple, TupleDesc tupleDescriptor, Oid interva
 /*
  * CachedRelationLookup performs a cached lookup for the relation
  * relationName, with the result cached in *cachedOid.
- *
- * NB: The way this is cached means the result will be wrong after the
- * extension is dropped and reconnect. A reconnect fixes that though, so that
- * seems acceptable.
  */
 static void
 CachedRelationLookup(const char *relationName, Oid *cachedOid)
