@@ -160,6 +160,9 @@ worker_merge_files_and_run_query(PG_FUNCTION_ARGS)
 
 	appendStringInfo(setSearchPathString, SET_SEARCH_PATH_COMMAND, jobSchemaName->data);
 
+	/* Add "public" to search path to access UDFs in public schema */
+	appendStringInfo(setSearchPathString, ",public");
+
 	connected = SPI_connect();
 	if (connected != SPI_OK_CONNECT)
 	{
@@ -365,7 +368,7 @@ CreateTaskTable(StringInfo schemaName, StringInfo relationName,
 	CreateStmt *createStatement = NULL;
 	RangeVar *relation = NULL;
 	List *columnDefinitionList = NIL;
-	Oid relationId = InvalidOid;
+	Oid relationId PG_USED_FOR_ASSERTS_ONLY = InvalidOid;
 #if (PG_VERSION_NUM >= 90500)
 	ObjectAddress relationObject;
 #endif
